"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
Object.defineProperty(exports, "parseKeyOfTuple", {
    enumerable: true,
    get: ()=>parseKeyOfTuple
});
const _propsTs = require("../../nodes/rules/props.js");
const _errorsTs = require("../../utils/errors.js");
const _freezeTs = require("../../utils/freeze.js");
const _genericsTs = require("../../utils/generics.js");
const _numericLiteralsTs = require("../../utils/numericLiterals.js");
const _objectKindsTs = require("../../utils/objectKinds.js");
const _serializeTs = require("../../utils/serialize.js");
const _definitionTs = require("../definition.js");
const _intersectionTs = require("./intersection.js");
var _keyNode, _keyType, _keyNode1, _keyNode2;
const arrayIndexStringBranch = (0, _freezeTs.deepFreeze)({
    regex: _numericLiteralsTs.wellFormedNonNegativeIntegerMatcher.source
});
const arrayIndexNumberBranch = (0, _freezeTs.deepFreeze)({
    range: {
        min: {
            comparator: ">=",
            limit: 1
        }
    },
    divisor: 1
});
const parseKeyOfTuple = (def, ctx)=>{
    const resolution = ctx.type.scope.resolveNode((0, _definitionTs.parseDefinition)(def[1], ctx));
    const predicateKeys = (0, _genericsTs.objectKeysOf)(resolution).map((domain)=>keysOfPredicate(domain, resolution[domain]));
    const sharedKeys = sharedKeysOf(predicateKeys);
    if (!sharedKeys.length) {
        return (0, _intersectionTs.writeImplicitNeverMessage)(ctx.path, "keyof");
    }
    const keyNode = {};
    for (const key of sharedKeys){
        const keyType = typeof key;
        if (keyType === "string" || keyType === "number" || keyType === "symbol") {
            (_keyNode = keyNode)[_keyType = keyType] ?? (_keyNode[_keyType] = []);
            keyNode[keyType].push({
                value: key
            });
        } else if (key === _numericLiteralsTs.wellFormedNonNegativeIntegerMatcher) {
            (_keyNode1 = keyNode).string ?? (_keyNode1.string = []);
            keyNode.string.push(arrayIndexStringBranch);
            (_keyNode2 = keyNode).number ?? (_keyNode2.number = []);
            keyNode.number.push(arrayIndexNumberBranch);
        } else {
            return (0, _errorsTs.throwInternalError)(`Unexpected keyof key '${(0, _serializeTs.stringify)(key)}'`);
        }
    }
    return Object.fromEntries(Object.entries(keyNode).map(([domain, branches])=>[
            domain,
            branches.length === 1 ? branches[0] : branches
        ]));
};
const baseKeysByDomain = {
    bigint: (0, _genericsTs.prototypeKeysOf)(0n),
    boolean: (0, _genericsTs.prototypeKeysOf)(false),
    null: [],
    number: (0, _genericsTs.prototypeKeysOf)(0),
    // TS doesn't include the Object prototype in keyof, so keyof object is never
    object: [],
    string: (0, _genericsTs.prototypeKeysOf)(""),
    symbol: (0, _genericsTs.prototypeKeysOf)(Symbol()),
    undefined: []
};
const keysOfPredicate = (domain, predicate)=>domain !== "object" || predicate === true ? baseKeysByDomain[domain] : sharedKeysOf((0, _genericsTs.listFrom)(predicate).map((branch)=>keysOfObjectBranch(branch)));
const sharedKeysOf = (keyBranches)=>{
    if (!keyBranches.length) {
        return [];
    }
    let sharedKeys = keyBranches[0];
    for(let i = 1; i < keyBranches.length; i++){
        // we can filter directly by equality here because the RegExp we're
        // using will always be reference equal to
        // wellFormedNonNegativeIntegerMatcher
        sharedKeys = sharedKeys.filter((k)=>keyBranches[i].includes(k));
    }
    return sharedKeys;
};
const keysOfObjectBranch = (branch)=>{
    const result = [];
    if ("props" in branch) {
        for (const key of Object.keys(branch.props)){
            if (key === _propsTs.mappedKeys.index) {
                result.push(_numericLiteralsTs.wellFormedNonNegativeIntegerMatcher);
            } else if (!result.includes(key)) {
                result.push(key);
                if (_numericLiteralsTs.wellFormedNonNegativeIntegerMatcher.test(key)) {
                    // allow numeric access to keys
                    result.push((0, _numericLiteralsTs.tryParseWellFormedInteger)(key, `Unexpectedly failed to parse an integer from key '${key}'`));
                }
            }
        }
    }
    if ("class" in branch) {
        const constructor = typeof branch.class === "string" ? _objectKindsTs.defaultObjectKinds[branch.class] : branch.class;
        for (const key of (0, _genericsTs.prototypeKeysOf)(constructor.prototype)){
            if (!result.includes(key)) {
                result.push(key);
            }
        }
    }
    return result;
};
