"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    parseDefinition: ()=>parseDefinition,
    as: ()=>as,
    unknownDefinitionMessage: ()=>unknownDefinitionMessage,
    writeBadDefinitionTypeMessage: ()=>writeBadDefinitionTypeMessage
});
const _typeTs = require("../scopes/type.js");
const _domainsTs = require("../utils/domains.js");
const _errorsTs = require("../utils/errors.js");
const _objectKindsTs = require("../utils/objectKinds.js");
const _serializeTs = require("../utils/serialize.js");
const _tupleTs = require("./ast/tuple.js");
const _recordTs = require("./record.js");
const _stringTs = require("./string/string.js");
const parseDefinition = (def, ctx)=>{
    const domain = (0, _domainsTs.domainOf)(def);
    if (domain === "string") {
        return (0, _stringTs.parseString)(def, ctx);
    }
    if (domain !== "object") {
        return (0, _errorsTs.throwParseError)(writeBadDefinitionTypeMessage(domain));
    }
    const objectKind = (0, _objectKindsTs.objectKindOf)(def);
    switch(objectKind){
        case "Object":
            return (0, _recordTs.parseRecord)(def, ctx);
        case "Array":
            return (0, _tupleTs.parseTuple)(def, ctx);
        case "RegExp":
            return {
                string: {
                    regex: def.source
                }
            };
        case "Function":
            if ((0, _typeTs.isType)(def)) {
                return ctx.type.scope.addAnonymousTypeReference(def, ctx);
            }
            if (isThunk(def)) {
                const returned = def();
                if ((0, _typeTs.isType)(returned)) {
                    return ctx.type.scope.addAnonymousTypeReference(returned, ctx);
                }
            }
            return (0, _errorsTs.throwParseError)(writeBadDefinitionTypeMessage("Function"));
        default:
            return (0, _errorsTs.throwParseError)(writeBadDefinitionTypeMessage(objectKind ?? (0, _serializeTs.stringify)(def)));
    }
};
const as = Symbol("as");
const unknownDefinitionMessage = "Cannot statically parse a definition inferred as unknown. Consider using 'as inferred<...>' to cast it.";
const isThunk = (def)=>typeof def === "function" && def.length === 0;
const writeBadDefinitionTypeMessage = (actual)=>`Type definitions must be strings or objects (was ${actual})`;
