"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    TraversalState: ()=>TraversalState,
    traverse: ()=>traverse,
    checkEntries: ()=>checkEntries,
    checkRequiredProp: ()=>checkRequiredProp
});
const _discriminateTs = require("../nodes/discriminate.js");
const _classTs = require("../nodes/rules/class.js");
const _divisorTs = require("../nodes/rules/divisor.js");
const _rangeTs = require("../nodes/rules/range.js");
const _regexTs = require("../nodes/rules/regex.js");
const _rulesTs = require("../nodes/rules/rules.js");
const _typeTs = require("../scopes/type.js");
const _domainsTs = require("../utils/domains.js");
const _errorsTs = require("../utils/errors.js");
const _genericsTs = require("../utils/generics.js");
const _pathsTs = require("../utils/paths.js");
const _problemsTs = require("./problems.js");
function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
const initializeTraversalConfig = ()=>({
        mustBe: [],
        writeReason: [],
        addContext: [],
        keys: []
    });
const problemWriterKeys = [
    "mustBe",
    "writeReason",
    "addContext"
];
var _seen = /*#__PURE__*/ new WeakMap();
class TraversalState {
    getProblemConfig(code) {
        const result = {};
        for (const k of problemWriterKeys){
            result[k] = this.traversalConfig[k][0] ?? this.rootScope.config.codes[code][k];
        }
        return result;
    }
    getConfigKey(k) {
        return this.traversalConfig[k][0];
    }
    traverseConfig(configEntries, node) {
        for (const entry of configEntries){
            this.traversalConfig[entry[0]].unshift(entry[1]);
        }
        const isValid = traverse(node, this);
        for (const entry of configEntries){
            this.traversalConfig[entry[0]].shift();
        }
        return isValid;
    }
    traverseKey(key, node) {
        const lastData = this.data;
        this.data = this.data[key];
        this.path.push(key);
        const isValid = traverse(node, this);
        this.path.pop();
        lastData[key] = this.data;
        this.data = lastData;
        return isValid;
    }
    traverseResolution(name) {
        const resolution = this.type.scope.resolve(name);
        const id = resolution.qualifiedName;
        // this assignment helps with narrowing
        const data = this.data;
        const isObject = (0, _domainsTs.hasDomain)(data, "object");
        if (isObject) {
            const seenByCurrentType = _classPrivateFieldGet(this, _seen)[id];
            if (seenByCurrentType) {
                if (seenByCurrentType.includes(data)) {
                    // if data has already been checked by this alias as part of
                    // a resolution higher up on the call stack, it must be valid
                    // or we wouldn't be here
                    return true;
                }
                seenByCurrentType.push(data);
            } else {
                _classPrivateFieldGet(this, _seen)[id] = [
                    data
                ];
            }
        }
        const lastType = this.type;
        this.type = resolution;
        const isValid = traverse(resolution.flat, this);
        this.type = lastType;
        if (isObject) {
            _classPrivateFieldGet(this, _seen)[id].pop();
        }
        return isValid;
    }
    traverseBranches(branches) {
        const lastFailFast = this.failFast;
        this.failFast = true;
        const lastProblems = this.problems;
        const branchProblems = new _problemsTs.Problems(this);
        this.problems = branchProblems;
        const lastPath = this.path;
        let hasValidBranch = false;
        for (const branch of branches){
            this.path = new _pathsTs.Path();
            if (checkEntries(branch, this)) {
                hasValidBranch = true;
                break;
            }
        }
        this.path = lastPath;
        this.problems = lastProblems;
        this.failFast = lastFailFast;
        return hasValidBranch || !this.problems.add("branches", branchProblems);
    }
    constructor(data, type){
        _classPrivateFieldInit(this, _seen, {
            writable: true,
            value: void 0
        });
        this.data = data;
        this.type = type;
        this.path = new _pathsTs.Path();
        this.problems = new _problemsTs.Problems(this);
        this.failFast = false;
        this.traversalConfig = initializeTraversalConfig();
        _classPrivateFieldSet(this, _seen, {});
        this.rootScope = type.scope;
    }
}
const traverse = (node, state)=>typeof node === "string" ? (0, _domainsTs.domainOf)(state.data) === node || !state.problems.add("domain", node) : checkEntries(node, state);
const checkEntries = (entries, state)=>{
    let isValid = true;
    for(let i = 0; i < entries.length; i++){
        const [k, v] = entries[i];
        const entryAllowsData = entryCheckers[k](v, state);
        isValid && (isValid = entryAllowsData);
        if (!isValid) {
            if (state.failFast) {
                return false;
            }
            if (i < entries.length - 1 && _rulesTs.precedenceMap[k] < _rulesTs.precedenceMap[entries[i + 1][0]]) {
                // if we've encountered a problem, there is at least one entry
                // remaining, and the next entry is of a higher precedence level
                // than the current entry, return immediately
                return false;
            }
        }
    }
    return isValid;
};
const checkRequiredProp = (prop, state)=>{
    if (prop[0] in state.data) {
        return state.traverseKey(prop[0], prop[1]);
    }
    state.problems.add("missing", undefined, {
        path: state.path.concat(prop[0]),
        data: undefined
    });
    return false;
};
const createPropChecker = (kind)=>(props, state)=>{
        let isValid = true;
        const unseenRequired = {
            ...props.required
        };
        for(const k in state.data){
            if (props.required[k]) {
                isValid || (isValid = state.traverseKey(k, props.required[k]));
                delete unseenRequired[k];
            } else if (props.optional[k]) {
                isValid || (isValid = state.traverseKey(k, props.optional[k]));
            } else if (kind === "distilledProps") {
                delete state.data[k];
            } else {
                state.problems.add("extraneous", state.data[k], {
                    path: state.path.concat(k)
                });
                isValid = false;
            }
            if (!isValid && state.failFast) {
                return false;
            }
        }
        return isValid;
    };
const entryCheckers = {
    regex: _regexTs.checkRegex,
    divisor: _divisorTs.checkDivisor,
    domains: (domains, state)=>{
        const entries = domains[(0, _domainsTs.domainOf)(state.data)];
        return entries ? checkEntries(entries, state) : !state.problems.add("cases", (0, _problemsTs.domainsToDescriptions)((0, _genericsTs.objectKeysOf)(domains)));
    },
    domain: (domain, state)=>(0, _domainsTs.domainOf)(state.data) === domain || !state.problems.add("domain", domain),
    bound: _rangeTs.checkBound,
    optionalProp: (prop, state)=>{
        if (prop[0] in state.data) {
            return state.traverseKey(prop[0], prop[1]);
        }
        return true;
    },
    // these checks work the same way, the keys are only distinct so that
    // prerequisite props can have a higher precedence
    requiredProp: checkRequiredProp,
    prerequisiteProp: checkRequiredProp,
    indexProp: (node, state)=>{
        if (!Array.isArray(state.data)) {
            state.problems.add("class", "Array");
            return false;
        }
        let isValid = true;
        for(let i = 0; i < state.data.length; i++){
            isValid && (isValid = state.traverseKey(`${i}`, node));
            if (!isValid && state.failFast) {
                return false;
            }
        }
        return isValid;
    },
    branches: (branches, state)=>state.traverseBranches(branches),
    switch: (rule, state)=>{
        const dataAtPath = (0, _pathsTs.getPath)(state.data, rule.path);
        const caseKey = (0, _discriminateTs.serializeCase)(rule.kind, dataAtPath);
        if ((0, _genericsTs.hasKey)(rule.cases, caseKey)) {
            return checkEntries(rule.cases[caseKey], state);
        }
        const caseKeys = (0, _genericsTs.objectKeysOf)(rule.cases);
        const missingCasePath = state.path.concat(rule.path);
        const caseDescriptions = rule.kind === "value" ? caseKeys : rule.kind === "domain" ? (0, _problemsTs.domainsToDescriptions)(caseKeys) : rule.kind === "class" ? (0, _problemsTs.objectKindsToDescriptions)(caseKeys) : (0, _errorsTs.throwInternalError)(`Unexpectedly encountered rule kind '${rule.kind}' during traversal`);
        state.problems.add("cases", caseDescriptions, {
            path: missingCasePath,
            data: dataAtPath
        });
        return false;
    },
    alias: (name, state)=>state.traverseResolution(name),
    class: _classTs.checkClass,
    narrow: (narrow, state)=>{
        const lastProblemsCount = state.problems.count;
        const result = narrow(state.data, state.problems);
        if (!result && state.problems.count === lastProblemsCount) {
            state.problems.mustBe(narrow.name ? `valid according to ${narrow.name}` : "valid");
        }
        return result;
    },
    config: ({ config , node  }, state)=>state.traverseConfig(config, node),
    value: (value, state)=>state.data === value || !state.problems.add("value", value),
    morph: (morph, state)=>{
        const out = morph(state.data, state.problems);
        if (state.problems.length) {
            return false;
        }
        if (out instanceof _problemsTs.Problem) {
            // if a problem was returned from the morph but not added, add it
            state.problems.addProblem(out);
            return false;
        }
        if ((0, _typeTs.isCheckResult)(out)) {
            if (out.problems) {
                for (const problem of out.problems){
                    state.problems.addProblem(problem);
                }
                return false;
            }
            state.data = out.data;
            return true;
        }
        state.data = out;
        return true;
    },
    distilledProps: createPropChecker("distilledProps"),
    strictProps: createPropChecker("strictProps")
};
