import type { distributable } from "../parse/ast/distributableFunction.js";
import type { Morph } from "../parse/ast/morph.js";
import type { Narrow } from "../parse/ast/narrow.js";
import type { FunctionalTupleOperator, inferTupleExpression, TupleExpression, TuplePostfixOperator, UnparsedTupleExpressionInput, UnparsedTupleOperator } from "../parse/ast/tuple.js";
import type { inferDefinition, validateDefinition } from "../parse/definition.js";
import type { conform } from "../utils/generics.js";
import type { PrecompiledDefaults } from "./ark.js";
import type { asIn, asOut, Type, TypeOptions } from "./type.js";
export type Expressions<$> = {
    intersection: BinaryExpressionParser<$, "&">;
    union: BinaryExpressionParser<$, "|">;
    arrayOf: UnaryExpressionParser<$, "[]">;
    keyOf: UnaryExpressionParser<$, "keyof">;
    node: UnvalidatedExpressionParser<$, "node">;
    instanceOf: UnvalidatedExpressionParser<$, "instanceof">;
    valueOf: UnvalidatedExpressionParser<$, "===">;
    narrow: FunctionalExpressionParser<$, "=>">;
    morph: FunctionalExpressionParser<$, "|>">;
};
type Ark = Expressions<PrecompiledDefaults>;
export declare const intersection: Ark["intersection"];
export declare const union: Ark["union"];
export declare const arrayOf: Ark["arrayOf"];
export declare const keyOf: Ark["keyOf"];
export declare const instanceOf: Ark["instanceOf"];
export declare const valueOf: Ark["valueOf"];
export declare const narrow: Ark["narrow"];
export declare const morph: Ark["morph"];
export type BinaryExpressionParser<$, operator extends "&" | "|"> = {
    <l, r>(l: validateDefinition<l, $>, r: validateDefinition<r, $>): parseTupleExpression<[l, operator, r], $>;
    <l, r>(l: validateDefinition<l, $>, r: validateDefinition<r, $>, opts: TypeOptions): parseTupleExpression<[l, operator, r], $>;
};
export type UnaryExpressionParser<$, operator extends "keyof" | "[]"> = {
    <def>(def: validateDefinition<def, $>): parseTupleExpression<unaryToTupleExpression<def, operator>, $>;
    <def>(def: validateDefinition<def, $>, opts: TypeOptions): parseTupleExpression<unaryToTupleExpression<def, operator>, $>;
};
export type UnvalidatedExpressionParser<$, operator extends UnparsedTupleOperator> = {
    <def>(def: conform<def, UnparsedTupleExpressionInput<$>[operator]>): parseTupleExpression<[operator, def], $>;
    <def>(def: conform<def, UnparsedTupleExpressionInput<$>[operator]>, opts: TypeOptions): parseTupleExpression<[operator, def], $>;
};
export type FunctionalExpressionParser<$, operator extends FunctionalTupleOperator> = {
    <inDef, fn extends FunctionWithInferredInput<$, operator, inDef>>(inDef: validateDefinition<inDef, $>, fn: fn): parseTupleExpression<[inDef, operator, fn], $>;
    <inDef, fn extends FunctionWithInferredInput<$, operator, inDef>>(inDef: validateDefinition<inDef, $>, fn: fn, opts: TypeOptions): parseTupleExpression<[inDef, operator, fn], $>;
};
export type FunctionWithInferredInput<$, operator extends FunctionalTupleOperator, inDef> = operator extends "=>" ? distributable<Narrow<asIn<inferDefinition<inDef, $>>>> : Morph<asOut<inferDefinition<inDef, $>>>;
type unaryToTupleExpression<def, operator extends "keyof" | "[]"> = operator extends TuplePostfixOperator ? [def, "[]"] : [operator, def];
type parseTupleExpression<expression extends TupleExpression, $> = inferTupleExpression<expression, $> extends infer result ? [result] extends [never] ? never : Type<result> : never;
export {};
